@@ -1,4 +1,4 @@
-/*	$OpenBSD: rde.h,v 1.317 2025/09/24 14:04:04 claudio Exp $ */
+/*	$OpenBSD: rde.h,v 1.337 2025/12/28 17:52:44 claudio Exp $ */
 
 /*
  * Copyright (c) 2003, 2004 Claudio Jeker <claudio@openbsd.org> and
@@ -26,6 +26,7 @@
 #include <stddef.h>
 
 #include "bgpd.h"
+#include "chash.h"
 #include "log.h"
 
 /* rde internal structures */
@@ -40,13 +41,19 @@ enum peer_state {
 LIST_HEAD(prefix_list, prefix);
 TAILQ_HEAD(prefix_queue, prefix);
 RB_HEAD(rib_tree, rib_entry);
+TAILQ_HEAD(rib_queue, rib_entry);
+LIST_HEAD(rib_pq_head, rib_pq);
 
 struct rib_entry {
 	RB_ENTRY(rib_entry)	 rib_e;
+	TAILQ_ENTRY(rib_entry)	 rib_queue;
 	struct prefix_queue	 prefix_h;
 	struct pt_entry		*prefix;
+	struct rib_pq_head	 rib_pq_list;
+	uint32_t		 pq_peer_id;
 	uint16_t		 rib_id;
-	uint16_t		 lock;
+	uint8_t			 lock;
+	uint8_t			 pq_mode;
 };
 
 struct rib {
@@ -71,8 +78,11 @@ struct rib {
  * Currently I assume that we can do that with the neighbor_ip...
  */
 RB_HEAD(peer_tree, rde_peer);
-RB_HEAD(prefix_tree, prefix);
-RB_HEAD(prefix_index, prefix);
+
+CH_HEAD(pend_prefix_hash, pend_prefix);
+TAILQ_HEAD(pend_prefix_queue, pend_prefix);
+CH_HEAD(pend_attr_hash, pend_prefix);
+TAILQ_HEAD(pend_attr_queue, pend_attr);
 
 struct rde_peer {
 	RB_ENTRY(rde_peer)		 entry;
@@ -83,12 +93,15 @@ struct rde_peer {
 	struct bgpd_addr		 local_v6_addr;
 	struct capabilities		 capa;
 	struct addpath_eval		 eval;
-	struct prefix_index		 adj_rib_out;
-	struct prefix_tree		 updates[AID_MAX];
-	struct prefix_tree		 withdraws[AID_MAX];
+	struct pend_attr_queue		 updates[AID_MAX];
+	struct pend_prefix_queue	 withdraws[AID_MAX];
+	struct pend_attr_hash		 pend_attrs;
+	struct pend_prefix_hash		 pend_prefixes;
 	struct filter_head		*out_rules;
 	struct ibufqueue		*ibufq;
+	struct rib_queue		 rib_pq_head;
 	monotime_t			 staletime[AID_MAX];
+	uint32_t			 adjout_bid;
 	uint32_t			 remote_bgpid;
 	uint32_t			 path_id_tx;
 	unsigned int			 local_if_scope;
@@ -161,7 +174,7 @@ enum attrtypes {
 #define ATTR_WELL_KNOWN		ATTR_TRANSITIVE
 
 struct attr {
-	RB_ENTRY(attr)			 entry;
+	uint64_t			 hash;
 	u_char				*data;
 	int				 refcnt;
 	uint16_t			 len;
@@ -170,17 +183,18 @@ struct attr {
 };
 
 struct rde_community {
-	RB_ENTRY(rde_community)		entry;
-	int				size;
-	int				nentries;
-	int				flags;
-	int				refcnt;
+	uint64_t			 hash;
+	int				 size;
+	int				 nentries;
+	int				 flags;
+	int				 refcnt;
 	struct community		*communities;
 };
 
 #define	PARTIAL_COMMUNITIES		0x01
 #define	PARTIAL_LARGE_COMMUNITIES	0x02
 #define	PARTIAL_EXT_COMMUNITIES		0x04
+#define	PARTIAL_DIRTY			0x08
 
 #define	F_ATTR_ORIGIN		0x00001
 #define	F_ATTR_ASPATH		0x00002
@@ -206,7 +220,7 @@ struct rde_community {
 #define DEFAULT_LPREF		100
 
 struct rde_aspath {
-	RB_ENTRY(rde_aspath)		 entry;
+	uint64_t			 hash;
 	struct attr			**others;
 	struct aspath			*aspath;
 	struct rde_aspa_state		 aspa_state;
@@ -221,6 +235,9 @@ struct rde_aspath {
 	uint8_t				 others_len;
 	uint8_t				 aspa_generation;
 };
+#define PATH_HASHOFF		offsetof(struct rde_aspath, med)
+#define PATH_HASHSTART(x)	((const uint8_t *)x + PATH_HASHOFF)
+#define PATH_HASHSIZE		(sizeof(struct rde_aspath) - PATH_HASHOFF)
 
 enum nexthop_state {
 	NEXTHOP_LOOKUP,
@@ -252,50 +269,41 @@ struct nexthop {
 #define NEXTHOP_CONNECTED	0x01
 };
 
+struct adjout_prefix;
+
 /* generic entry without address specific part */
 struct pt_entry {
 	RB_ENTRY(pt_entry)		 pt_e;
+	struct adjout_prefix		*adjout;
+	uint32_t			 adjoutlen;
+	uint32_t			 adjoutavail;
 	uint8_t				 aid;
 	uint8_t				 prefixlen;
 	uint16_t			 len;
 	uint32_t			 refcnt;
-	uint8_t				 data[4]; /* data depending on aid */
+	uint8_t				 data[0]; /* data depending on aid */
 };
 
 struct prefix {
-	union {
-		struct {
-			TAILQ_ENTRY(prefix)	 rib;
-			LIST_ENTRY(prefix)	 nexthop;
-			struct rib_entry	*re;
-		} list;
-		struct {
-			RB_ENTRY(prefix)	 index, update;
-		} tree;
-	}				 entry;
-	struct pt_entry			*pt;
-	struct rde_aspath		*aspath;
-	struct rde_community		*communities;
-	struct rde_peer			*peer;
-	struct nexthop			*nexthop;	/* may be NULL */
-	monotime_t			 lastchange;
-	uint32_t			 path_id;
-	uint32_t			 path_id_tx;
-	uint16_t			 flags;
-	uint8_t				 validation_state;
-	uint8_t				 nhflags;
-	int8_t				 dmetric;	/* decision metric */
+	TAILQ_ENTRY(prefix)	 rib_l;
+	LIST_ENTRY(prefix)	 nexthop_l;
+	struct rib_entry	*re;
+	struct pt_entry		*pt;
+	struct rde_aspath	*aspath;
+	struct rde_community	*communities;
+	struct rde_peer		*peer;
+	struct nexthop		*nexthop;	/* may be NULL */
+	monotime_t		 lastchange;
+	uint32_t		 path_id;
+	uint32_t		 path_id_tx;
+	uint8_t			 flags;
+	uint8_t			 validation_state;
+	uint8_t			 nhflags;
+	int8_t			 dmetric;	/* decision metric */
 };
-#define	PREFIX_FLAG_WITHDRAW	0x0001	/* enqueued on withdraw queue */
-#define	PREFIX_FLAG_UPDATE	0x0002	/* enqueued on update queue */
-#define	PREFIX_FLAG_DEAD	0x0004	/* locked but removed */
-#define	PREFIX_FLAG_STALE	0x0008	/* stale entry (for addpath) */
-#define	PREFIX_FLAG_MASK	0x000f	/* mask for the prefix types */
-#define	PREFIX_FLAG_ADJOUT	0x0010	/* prefix is in the adj-out rib */
-#define	PREFIX_FLAG_EOR		0x0020	/* prefix is EoR */
-#define	PREFIX_NEXTHOP_LINKED	0x0040	/* prefix is linked onto nexthop list */
-#define	PREFIX_FLAG_LOCKED	0x0080	/* locked by rib walker */
-#define	PREFIX_FLAG_FILTERED	0x0100	/* prefix is filtered (ineligible) */
+
+#define	PREFIX_NEXTHOP_LINKED	0x01	/* prefix is linked onto nexthop list */
+#define	PREFIX_FLAG_FILTERED	0x04	/* prefix is filtered (ineligible) */
 
 #define	PREFIX_DMETRIC_NONE	0
 #define	PREFIX_DMETRIC_INVALID	1
@@ -312,6 +320,34 @@ struct prefix {
 #define	NEXTHOP_MASK		0x0f
 #define	NEXTHOP_VALID		0x80
 
+struct adjout_attr {
+	uint64_t		 hash;
+	struct rde_aspath	*aspath;
+	struct rde_community	*communities;
+	struct nexthop		*nexthop;
+	int			 refcnt;
+};
+
+struct adjout_prefix {
+	uint32_t			 path_id_tx;
+	struct adjout_attr		*attrs;
+	struct bitmap			 peermap;
+};
+
+struct pend_attr {
+	TAILQ_ENTRY(pend_attr)		 entry;
+	struct pend_prefix_queue	 prefixes;
+	struct adjout_attr		*attrs;
+	uint8_t				 aid;
+};
+
+struct pend_prefix {
+	TAILQ_ENTRY(pend_prefix)	 entry;
+	struct pt_entry			*pt;
+	struct pend_attr		*attrs;
+	uint32_t			 path_id_tx;
+};
+
 struct filterstate {
 	struct rde_aspath	 aspath;
 	struct rde_community	 communities;
@@ -321,9 +357,27 @@ struct filterstate {
 };
 
 enum eval_mode {
+	EVAL_NONE,
+	EVAL_RECONF,
 	EVAL_DEFAULT,
 	EVAL_ALL,
-	EVAL_RECONF,
+};
+
+struct rib_context {
+	LIST_ENTRY(rib_context)		 entry;
+	struct rib_entry		*ctx_re;
+	struct pt_entry			*ctx_pt;
+	uint32_t			 ctx_id;
+	void		(*ctx_rib_call)(struct rib_entry *, void *);
+	void		(*ctx_prefix_call)(struct rde_peer *,
+			    struct pt_entry *, struct adjout_prefix *, void *);
+	void		(*ctx_done)(void *, uint8_t);
+	int		(*ctx_throttle)(void *);
+	void				*ctx_arg;
+	struct bgpd_addr		 ctx_subtree;
+	unsigned int			 ctx_count;
+	uint8_t				 ctx_aid;
+	uint8_t				 ctx_subtreelen;
 };
 
 extern struct rde_memstats rdemem;
@@ -367,7 +421,8 @@ struct filter_head	*peer_apply_out_filter(struct rde_peer *,
 			    struct filter_head *);
 
 void		 rde_generate_updates(struct rib_entry *, struct prefix *,
-		    struct prefix *, enum eval_mode);
+		    uint32_t, enum eval_mode);
+void		 peer_process_updates(struct rde_peer *, void *);
 
 void		 peer_up(struct rde_peer *, struct session_up *);
 void		 peer_down(struct rde_peer *);
@@ -378,7 +433,6 @@ void		 peer_blast(struct rde_peer *, uint8_t);
 void		 peer_dump(struct rde_peer *, uint8_t);
 void		 peer_begin_rrefresh(struct rde_peer *, uint8_t);
 int		 peer_work_pending(void);
-void		 peer_reaper(struct rde_peer *);
 
 void		 peer_imsg_push(struct rde_peer *, struct imsg *);
 int		 peer_imsg_pop(struct rde_peer *, struct imsg *);
@@ -395,12 +449,13 @@ RB_PROTOTYPE(peer_tree, rde_peer, entry, peer_cmp);
 /* rde_attr.c */
 int		 attr_writebuf(struct ibuf *, uint8_t, uint8_t, void *,
 		    uint16_t);
-void		 attr_shutdown(void);
+void		 attr_init(void);
 int		 attr_optadd(struct rde_aspath *, uint8_t, uint8_t,
 		    void *, uint16_t);
 struct attr	*attr_optget(const struct rde_aspath *, uint8_t);
 void		 attr_copy(struct rde_aspath *, const struct rde_aspath *);
-int		 attr_compare(struct rde_aspath *, struct rde_aspath *);
+int		 attr_equal(const struct rde_aspath *,
+		    const struct rde_aspath *);
 void		 attr_freeall(struct rde_aspath *);
 void		 attr_free(struct rde_aspath *, struct attr *);
 
@@ -411,7 +466,7 @@ u_char		*aspath_deflate(u_char *, uint16_t *, int *);
 void		 aspath_merge(struct rde_aspath *, struct attr *);
 uint32_t	 aspath_neighbor(struct aspath *);
 int		 aspath_loopfree(struct aspath *, uint32_t);
-int		 aspath_compare(struct aspath *, struct aspath *);
+int		 aspath_compare(const struct aspath *, const struct aspath *);
 int		 aspath_match(struct aspath *, struct filter_as *, uint32_t);
 u_char		*aspath_prepend(struct aspath *, uint32_t, int, uint16_t *);
 u_char		*aspath_override(struct aspath *, uint32_t, uint32_t,
@@ -450,12 +505,13 @@ int	community_large_add(struct rde_community *, int, struct ibuf *);
 int	community_ext_add(struct rde_community *, int, int, struct ibuf *);
 int	community_writebuf(struct rde_community *, uint8_t, int, struct ibuf *);
 
-void			 communities_shutdown(void);
+void			 communities_init(void);
 struct rde_community	*communities_lookup(struct rde_community *);
 struct rde_community	*communities_link(struct rde_community *);
 void			 communities_unlink(struct rde_community *);
 
-int	 communities_equal(struct rde_community *, struct rde_community *);
+int	 communities_equal(const struct rde_community *,
+	    const struct rde_community *);
 void	 communities_copy(struct rde_community *, struct rde_community *);
 void	 communities_clean(struct rde_community *);
 
@@ -511,9 +567,12 @@ void	 pt_getaddr(struct pt_entry *, struct bgpd_addr *);
 int	 pt_getflowspec(struct pt_entry *, uint8_t **);
 struct pt_entry	*pt_fill(struct bgpd_addr *, int);
 struct pt_entry	*pt_get(struct bgpd_addr *, int);
-struct pt_entry *pt_add(struct bgpd_addr *, int);
+struct pt_entry	*pt_get_next(struct bgpd_addr *, int);
+struct pt_entry	*pt_add(struct bgpd_addr *, int);
 struct pt_entry	*pt_get_flow(struct flowspec *);
 struct pt_entry	*pt_add_flow(struct flowspec *);
+struct pt_entry	*pt_first(uint8_t);
+struct pt_entry	*pt_next(struct pt_entry *);
 void	 pt_remove(struct pt_entry *);
 struct pt_entry	*pt_lookup(struct bgpd_addr *);
 int	 pt_prefix_cmp(const struct pt_entry *, const struct pt_entry *);
@@ -551,6 +610,7 @@ struct rib_entry *rib_get_addr(struct rib *, struct bgpd_addr *, int);
 struct rib_entry *rib_match(struct rib *, struct bgpd_addr *);
 int		 rib_dump_pending(void);
 void		 rib_dump_runner(void);
+void		 rib_dump_insert(struct rib_context *);
 int		 rib_dump_new(uint16_t, uint8_t, unsigned int, void *,
 		    void (*)(struct rib_entry *, void *),
 		    void (*)(void *, uint8_t),
@@ -561,6 +621,7 @@ int		 rib_dump_subtree(uint16_t, struct bgpd_addr *, uint8_t,
 		    void (*)(void *, uint8_t),
 		    int (*)(void *));
 void		 rib_dump_terminate(void *);
+void		 rib_dequeue(struct rib_entry *);
 
 extern struct rib flowrib;
 
@@ -573,7 +634,12 @@ re_rib(struct rib_entry *re)
 	return rib_byid(re->rib_id);
 }
 
-void		 path_shutdown(void);
+void		 path_init(void);
+struct rde_aspath *path_ref(struct rde_aspath *);
+void		 path_unref(struct rde_aspath *);
+int		 path_equal(const struct rde_aspath *,
+		    const struct rde_aspath *);
+struct rde_aspath *path_getcache(struct rde_aspath *);
 struct rde_aspath *path_copy(struct rde_aspath *, const struct rde_aspath *);
 struct rde_aspath *path_prep(struct rde_aspath *);
 struct rde_aspath *path_get(void);
@@ -583,13 +649,6 @@ void		 path_put(struct rde_aspath *);
 #define	PREFIX_SIZE(x)	(((x) + 7) / 8 + 1)
 struct prefix	*prefix_get(struct rib *, struct rde_peer *, uint32_t,
 		    struct bgpd_addr *, int);
-struct prefix	*prefix_adjout_get(struct rde_peer *, uint32_t,
-		    struct pt_entry *);
-struct prefix	*prefix_adjout_first(struct rde_peer *, struct pt_entry *);
-struct prefix	*prefix_adjout_next(struct rde_peer *, struct prefix *);
-struct prefix	*prefix_adjout_lookup(struct rde_peer *, struct bgpd_addr *,
-		    int);
-struct prefix	*prefix_adjout_match(struct rde_peer *, struct bgpd_addr *);
 int		 prefix_update(struct rib *, struct rde_peer *, uint32_t,
 		    uint32_t, struct filterstate *, int, struct bgpd_addr *,
 		    int);
@@ -601,26 +660,11 @@ int		 prefix_flowspec_withdraw(struct rde_peer *, struct pt_entry *);
 void		 prefix_flowspec_dump(uint8_t, void *,
 		    void (*)(struct rib_entry *, void *),
 		    void (*)(void *, uint8_t));
-void		 prefix_add_eor(struct rde_peer *, uint8_t);
-void		 prefix_adjout_update(struct prefix *, struct rde_peer *,
-		    struct filterstate *, struct pt_entry *, uint32_t);
-void		 prefix_adjout_withdraw(struct prefix *);
-void		 prefix_adjout_destroy(struct prefix *);
-void		 prefix_adjout_flush_pending(struct rde_peer *);
-int		 prefix_adjout_reaper(struct rde_peer *);
-int		 prefix_dump_new(struct rde_peer *, uint8_t, unsigned int,
-		    void *, void (*)(struct prefix *, void *),
-		    void (*)(void *, uint8_t), int (*)(void *));
-int		 prefix_dump_subtree(struct rde_peer *, struct bgpd_addr *,
-		    uint8_t, unsigned int, void *,
-		    void (*)(struct prefix *, void *),
-		    void (*)(void *, uint8_t), int (*)(void *));
+
 struct prefix	*prefix_bypeer(struct rib_entry *, struct rde_peer *,
 		    uint32_t);
 void		 prefix_destroy(struct prefix *);
 
-RB_PROTOTYPE(prefix_tree, prefix, entry, prefix_cmp)
-
 static inline struct rde_peer *
 prefix_peer(struct prefix *p)
 {
@@ -679,9 +723,7 @@ prefix_set_vstate(struct prefix *p, uint8_t roa_vstate, uint8_t aspa_vstate)
 static inline struct rib_entry *
 prefix_re(struct prefix *p)
 {
-	if (p->flags & PREFIX_FLAG_ADJOUT)
-		return NULL;
-	return (p->entry.list.re);
+	return (p->re);
 }
 
 static inline int
@@ -702,11 +744,49 @@ struct nexthop	*nexthop_get(struct bgpd_addr *);
 struct nexthop	*nexthop_ref(struct nexthop *);
 int		 nexthop_unref(struct nexthop *);
 
+/* rde_adjout.c */
+void			 adjout_init(void);
+struct adjout_prefix	*adjout_prefix_get(struct rde_peer *, uint32_t,
+			    struct pt_entry *);
+struct adjout_prefix	*adjout_prefix_first(struct rde_peer *,
+			    struct pt_entry *);
+struct adjout_prefix	*adjout_prefix_next(struct rde_peer *,
+			    struct pt_entry *, struct adjout_prefix *);
+
+void		 prefix_add_eor(struct rde_peer *, uint8_t);
+void		 adjout_prefix_update(struct adjout_prefix *, struct rde_peer *,
+		    struct filterstate *, struct pt_entry *, uint32_t);
+void		 adjout_prefix_withdraw(struct rde_peer *, struct pt_entry *,
+		    struct adjout_prefix *);
+void		 adjout_prefix_reaper(struct rde_peer *);
+void		 adjout_prefix_dump_cleanup(struct rib_context *);
+void		 adjout_prefix_dump_r(struct rib_context *);
+int		 adjout_prefix_dump_new(struct rde_peer *, uint8_t,
+		    unsigned int, void *,
+		    void (*)(struct rde_peer *, struct pt_entry *,
+			struct adjout_prefix *, void *),
+		    void (*)(void *, uint8_t), int (*)(void *));
+int		 adjout_prefix_dump_subtree(struct rde_peer *,
+		    struct bgpd_addr *, uint8_t, unsigned int, void *,
+		    void (*)(struct rde_peer *, struct pt_entry *,
+			struct adjout_prefix *, void *),
+		    void (*)(void *, uint8_t), int (*)(void *));
+void		 adjout_peer_init(struct rde_peer *);
+void		 adjout_peer_flush_pending(struct rde_peer *);
+void		 adjout_peer_free(struct rde_peer *);
+
+void		 pend_attr_done(struct pend_attr *, struct rde_peer *);
+void		 pend_eor_add(struct rde_peer *, uint8_t);
+void		 pend_prefix_add(struct rde_peer *, struct adjout_attr *,
+		    struct pt_entry *, uint32_t);
+void		 pend_prefix_free(struct pend_prefix *,
+		    struct pend_prefix_queue *, struct rde_peer *);
+
 /* rde_update.c */
 void		 up_generate_updates(struct rde_peer *, struct rib_entry *);
 void		 up_generate_addpath(struct rde_peer *, struct rib_entry *);
 void		 up_generate_addpath_all(struct rde_peer *, struct rib_entry *,
-		    struct prefix *, struct prefix *);
+		    struct prefix *, uint32_t);
 void		 up_generate_default(struct rde_peer *, uint8_t);
 int		 up_is_eor(struct rde_peer *, uint8_t);
 void		 up_dump_withdraws(struct imsgbuf *, struct rde_peer *,
